import enum
import json
import typing

from renus.core.serialize import jsonEncoder
from renus.core.injection import Injection
from renus.core.request import Request


class WebSocketState(enum.Enum):
    CONNECTING = 0
    CONNECTED = 1
    DISCONNECTED = 2


class WebSocketDisconnect(Exception):
    def __init__(self, code: int = 1000) -> None:
        self.code = code


class WebSocket(Request):
    def __init__(self, scope, receive, send) -> None:
        super().__init__(scope, receive)
        assert scope["type"] == "websocket"
        self._receive = receive
        self._scope = scope
        self._send = send
        self.client_state = WebSocketState.CONNECTING
        self.application_state = WebSocketState.CONNECTING

    async def receive(self):
        """
        Receive ASGI websocket messages, ensuring valid state transitions.
        """
        if self.client_state == WebSocketState.CONNECTING:
            message = await self._receive()
            message_type = message["type"]
            assert message_type == "websocket.connect"
            self.client_state = WebSocketState.CONNECTED
            return message
        elif self.client_state == WebSocketState.CONNECTED:
            message = await self._receive()
            message_type = message["type"]
            assert message_type in {"websocket.receive", "websocket.disconnect"}
            if message_type == "websocket.disconnect":
                self.client_state = WebSocketState.DISCONNECTED
            return message
        else:
            raise RuntimeError(
                'Cannot call "receive" once a disconnect message has been received.'
            )

    async def send(self, message) -> None:
        """
        Send ASGI websocket messages, ensuring valid state transitions.
        """
        if self.application_state == WebSocketState.CONNECTING:
            message_type = message["type"]
            assert message_type in {"websocket.accept", "websocket.close"}
            if message_type == "websocket.close":
                self.application_state = WebSocketState.DISCONNECTED
            else:
                self.application_state = WebSocketState.CONNECTED
            await self._send(message)
        elif self.application_state == WebSocketState.CONNECTED:
            message_type = message["type"]
            assert message_type in {"websocket.send", "websocket.close"}
            if message_type == "websocket.close":
                self.application_state = WebSocketState.DISCONNECTED
            await self._send(message)
        else:
            raise RuntimeError('Cannot call "send" once a close message has been sent.')

    async def accept(self, subprotocol: str = None) -> None:
        if self.client_state == WebSocketState.CONNECTING:
            # If we haven't yet seen the 'connect' message, then wait for it first.
            await self.receive()
        await self.send({"type": "websocket.accept", "subprotocol": subprotocol})

    def _raise_on_disconnect(self, message) -> None:
        if message["type"] == "websocket.disconnect":
            raise WebSocketDisconnect(message["code"])

    async def receive_text(self,protect:bool=True) -> str:
        assert self.application_state == WebSocketState.CONNECTED
        message = await self.receive()
        self._raise_on_disconnect(message)
        return Injection().protect(message["text"]) if protect else message["text"]

    async def receive_bytes(self) -> bytes:
        assert self.application_state == WebSocketState.CONNECTED
        message = await self.receive()
        self._raise_on_disconnect(message)
        return message["bytes"]

    async def receive_json(self,protect:bool=True) -> typing.Any:
        assert self.application_state == WebSocketState.CONNECTED
        message = await self.receive()
        self._raise_on_disconnect(message)
        if "text" in message:
            text = message["text"]
        else:
            text = message["bytes"].decode("utf-8")

        return Injection().protect(json.loads(text)) if protect else json.loads(text)

    async def iter_text(self) -> typing.AsyncIterator[str]:
        try:
            while True:
                yield await self.receive_text()
        except WebSocketDisconnect:
            pass

    async def iter_bytes(self) -> typing.AsyncIterator[bytes]:
        try:
            while True:
                yield await self.receive_bytes()
        except WebSocketDisconnect:
            pass

    async def iter_json(self) -> typing.AsyncIterator[typing.Any]:
        try:
            while True:
                yield await self.receive_json()
        except WebSocketDisconnect:
            pass

    async def send_text(self, data: str) -> None:
        await self.send({"type": "websocket.send", "text": data})

    async def send_bytes(self, data: bytes) -> None:
        await self.send({"type": "websocket.send", "bytes": data})

    async def send_json(self, data: typing.Any, mode: str = "text") -> None:
        assert mode in ["text", "binary"]
        text = json.dumps(data,
                          ensure_ascii=False,
                          allow_nan=True,
                          indent=None,
                          separators=(",", ":"),
                          cls=jsonEncoder)
        if mode == "text":
            await self.send({"type": "websocket.send", "text": text})
        else:
            await self.send({"type": "websocket.send", "bytes": text.encode("utf-8")})

    async def close(self, code: int = 1000) -> None:
        try:
            await self.send({"type": "websocket.close", "code": code})
        except Exception:
            pass



from json import dumps as json_dump, loads as json_loads
from random import choice as random_choice
from base64 import urlsafe_b64encode, urlsafe_b64decode
from hashlib import sha3_512
from renus.core.serialize import jsonEncoder

def hash_new_password(password: str, salt: str = 'renus') -> str:
    p = password.encode("utf-8")
    s = salt.encode("utf-8")
    r = sha3_512(s + p)
    return sha3_512(p + r.digest()).hexdigest()


def is_correct_password(pw_hash: bytes, password: str, salt: str = 'renus') -> bool:
    p = password.encode("utf-8")
    s = salt.encode("utf-8")
    r = sha3_512(s + p)
    return sha3_512(p + r.digest()).hexdigest() == pw_hash


def get_random_string(length):
    letters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
    return ''.join(random_choice(letters) for i in range(length))


def encode64(value):
    base_bytes = urlsafe_b64encode(json_dump(value,
                                             ensure_ascii=False,
                                             allow_nan=True,
                                             indent=4,
                                             separators=(",", ":"),
                                             cls=jsonEncoder).encode('utf-8'))
    return remove_pad64(base_bytes.decode())


def decode64(value: str, decoder=None):
    try:
        value = add_pad64(value)
        message_bytes = urlsafe_b64decode(value)
        if decoder is None:
            return json_loads(message_bytes)
        return json_loads(message_bytes, object_hook=decoder)
    except:
        raise RuntimeError('Invalid encode64 Format:' + str(value))



class dictAttribute(dict):
    def __init__(self, seq=None, **kwargs):
        super().__init__(seq, **kwargs)

    def __getattr__(self, key):
        ret = self.get(key)
        if not ret and key.startswith("__"):
            raise AttributeError()
        if type(ret) == dict:
            return dictAttribute(ret)
        return ret

    def __setattr__(self, key, value):
        self[key] = value

    def copy(self):
        return dictAttribute(dict(self).copy())


def remove_pad64(text):
    return text.replace('=', '')


def add_pad64(text):
    s = 4 - len(text) % 4
    if s == 4:
        return text
    return text + s * '='



import shutil
from io import BytesIO
import zipfile
from renus.commands.help import bc
from renus.commands.install.service import request

from renus.commands.install.build import _add_route, _add_admin_templates, _add_index_templates, _add_user_templates, \
    _add_db, _add_cls, _add_imprt


def download(app, version, installed, isUpdate=False):
    if app in installed:
        if isUpdate and installed[app] == version:
            return version
        elif not isUpdate:
            return version

    print(f'{bc.OKGREEN}downloading {app}/{version}{bc.ENDC}')

    res = request(f'https://reapp.codenus.com/app/{app}/{version}')

    if res.status_code != 200:
        print(f'{bc.FAIL}download error{bc.ENDC}')
        return

    if app in installed:
        if isUpdate and installed[app] == res.headers['version']:
            print('   same version', res.headers['version'])
            return res.headers['version']
        elif not isUpdate:
            print('   currently installed version', installed[app])
            return res.headers['version']

    installed[app] = res.headers['version']
    f = zipfile.ZipFile(BytesIO(res.content))
    shutil.rmtree(f"extension/{app}", True)
    f.extractall(f"extension/{app}")
    if not isUpdate:
        install_app(app, installed)
    else:
        print(f'{bc.OKBLUE}   updated to version{res.headers["version"]}{bc.ENDC}')
    return res.headers['version']


def rewrite(installed):
    print('   rewrite installed')
    pre_line = ''
    with open('extension/install.py', 'r') as f:
        lines = f.readlines()
        for line in lines:
            if line.find('installed') != -1:
                break
            pre_line += line
    with open('extension/install.py', 'w') as f:
        f.write(pre_line)
        f.write('installed = {\n')
        for app, version in installed.items():
            f.write(f'  "{app}":"{version}",\n')
        f.write('}')


def install_app(app, installed):
    print(f'   install {app}')
    try:
        install = __import__(f"extension.{app.replace('/', '.')}.install", fromlist=[''])

    except Exception as exc:
        print('   ' + bc.WARNING + str(exc) + bc.ENDC)
        return

    if hasattr(install, 'dependencies'):
        print(f'   install {app} dependencies...')
        for dep in install.dependencies:
            download(dep, '*', installed)

    if hasattr(install, 'cls'):
        print(f'   add {app} class...')
        _add_cls(install,app)

    if hasattr(install, 'route'):
        print(f'   add {app} route...')
        _add_route(install,app)

    if hasattr(install, 'imprt'):
        print(f'   add {app} imports...')
        _add_imprt(install,app)

    if hasattr(install, 'admin_templates'):
        print(f'   add {app} admin_templates...')
        _add_admin_templates(install)

    if hasattr(install, 'index_templates'):
        print(f'   add {app} index_templates...')
        _add_index_templates(install)

    if hasattr(install, 'user_templates'):
        print(f'   add {app} user_templates...')
        _add_user_templates(install)

    if hasattr(install, 'db'):
        print(f'   add {app} DB...')
        _add_db(install, app)


def install_all(apps, installed, isUpdate):
    for app, version in apps.items():
        download(app, version, installed, isUpdate)


def run(args=None):
    print('start...')
    try:
        from extension.install import apps, installed
    except:
        apps = {}
        installed = {}
    isUpdate = False

    if args and '-update' in args:
        isUpdate = True
        args.remove('-update')

    if args is None or len(args) == 0:
        install_all(apps, installed, isUpdate)
        pass
    else:
        s = args[0].split('/')
        version = s[-1]
        s.pop(-1)

        download('/'.join(s), version, installed, isUpdate)
    rewrite(installed)

    print(f'{bc.OKBLUE}finished{bc.ENDC}')



from base64 import urlsafe_b64encode, urlsafe_b64decode
import sys

from renus.util.helper import get_random_string, remove_pad64, add_pad64

try:
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
    from cryptography.fernet import Fernet
except ImportError:
    Fernet = None

def pad(s,n=16):
    block_size = n
    remainder = len(s) % block_size
    padding_needed = block_size - remainder
    return s + padding_needed * ' '


def __generate_key(password):
    salt = b'dfkTGG536fdfkTGG536fkdfijdkdfijd'
    kdf = PBKDF2HMAC(algorithm = hashes.SHA256(),length = 32,salt = salt,iterations = 100000)
    return urlsafe_b64encode(kdf.derive(password.encode()))


def encrypt(text:str,password:str)->str:
    assert Fernet is not None, "'cryptography' must be installed to use encrypt"
    key = __generate_key(password)
    f = Fernet(key)
    d= f.encrypt(text.encode())
    return d.decode()



def decrypt(enc_dict:str, password:str)->str:
    assert Fernet is not None, "'cryptography' must be installed to use decrypt"
    key = __generate_key(password)
    f = Fernet(key)
    d = f.decrypt(enc_dict.encode())
    return d.decode()


def shift(clear, key):
    clear += key
    rnd = get_random_string(6)
    clear = rnd + clear
    clear = urlsafe_b64encode(clear.encode()).decode()
    enc = []
    lnk=len(key)
    for i in range(len(clear)):
        key_c = key[i % lnk]
        enc_c = chr((ord(clear[i]) + ord(key_c)) % 256)
        enc.append(enc_c)

    return remove_pad64(urlsafe_b64encode("".join(enc).encode()).decode())


def unshift(enc, key):
    try:
        enc = urlsafe_b64decode(add_pad64(enc)).decode()
        ln = len(key)
        dec = []
        for i in range(len(enc)):
            key_c = key[i % ln]
            dec_c = chr((256 + ord(enc[i]) - ord(key_c)) % 256)
            dec.append(dec_c)
        return urlsafe_b64decode("".join(dec)).decode()[6:-ln]
    except:
        raise RuntimeError('Invalid Token Format')


from html.parser import HTMLParser
from html import unescape

class Injection:
    def escape(self, value:str):
        parser = Xss()
        parser.feed(unescape(value))
        parser.close()
        return parser.getEscaped().strip()

    def protect(self, data):
        return self.__value_handle(data)

    def __list_handle(self, data):
        res = []
        for item in data:
            res.append(self.__value_handle(item))
        return res

    def __value_handle(self, value):
        typ = type(value)
        if value is None or typ in [bool, int, float]:
            v = value
        elif typ is dict:
            v = self.__dict_handle(value)
        elif typ is list:
            v = self.__list_handle(value)
        else:
            v = self.escape(str(value))
        return v

    def __dict_handle(self, obj: dict):
        res = {}
        for key, value in obj.items():
            res[self.escape(str(key))] = self.__value_handle(value)
        return res


class Xss(HTMLParser):
    block_tags = {
        'script': 'scrlpt',
    }
    allow_protocol=['http','https']
    allow_attr = {'class':1,
                  'id':1,
                  'href':['a'],
                  'target':['a'],
                  'alt':['img'],
                  'width':['img','video'],
                  'height':['img','video'],
                  'controls':['video'],
                  'src':['img','video']
                  }
    no_end_tags = ["img", "hr", "br", "embed"]

    def __init__(self):
        HTMLParser.__init__(self)
        self.result = []
        self.start_list = []

    def getEscaped(self):
        return self._htmlspecialchars(''.join(self.result))

    def handle_startendtag(self, tag, attrs):
        self.handle_starttag(tag, attrs)

    def handle_starttag(self, tag, attrs):
        if tag in self.block_tags:
            tag = self.block_tags[tag]
        end_diagonal = ' /' if tag in self.no_end_tags else ''
        if not end_diagonal:
            self.start_list.append(tag)
        attdict = {}
        for attr in attrs:
            if attr[0] in self.allow_attr:
                if self.allow_attr[attr[0]]==1 or tag in self.allow_attr[attr[0]]:
                    attdict[attr[0]] =self._is_valid(attr)

        attrs = []
        for (key, value) in attdict.items():
            attrs.append('%s="%s"' % (key, self._htmlspecialchars(value)))
        attrs = (' ' + ' '.join(attrs)) if attrs else ''
        self.result.append('&lt;' + tag + attrs + end_diagonal + '&gt;')

    def handle_endtag(self, tag):
        if tag in self.block_tags:
            tag = self.block_tags[tag]
        if self.start_list and tag == self.start_list[len(self.start_list) - 1]:
            self.result.append('&lt;/' + tag + '&gt;')
            self.start_list.pop()

    def handle_data(self, data):
        self.result.append(self._htmlspecialchars(data))

    def handle_entityref(self, name):
        if name.isalpha():
            self.result.append("&%s;" % name)

    def handle_charref(self, name):
        if name.isdigit():
            self.result.append("&#%s;" % name)

    def _is_valid(self,attr):
        if attr[0] in ['href','src']:
            if str(attr[1]).find(':') !=-1:
                has =False
                for item in self.allow_protocol:
                    if str(attr[1]).startswith(item):
                        has=True
                if has is False:
                    return 'http://%s' % str(attr[1])
            elif str(attr[1]).startswith('//'):
                return 'http:%s' % str(attr[1])
        return attr[1]
    def _htmlspecialchars(self, html):
        return html.replace("<", "&lt;") \
            .replace(">", "&gt;") \
            .replace('"', "&quot;") \
            .replace("'", "&#039;")



import os
import zipfile

exclude_all = [
    '__pycache__',
    'frontend',
    '.idea',
    '.git',
    '.github'
]
exclude = [
    'venv',
    'storage/cache',
    'storage/logs'
]

exclude_ext = [
    '.vue',
    '.js'
    '.css'
    '.scss'
]


def zipdir(path, ziph):
    for dirname, subdirs, files in os.walk(path):
        has = False
        for dir in exclude_all:
            if dirname.find(dir) != -1:
                has = True
                break
        for dir in exclude:
            if dirname.startswith(dir):
                has = True
                break
        if has:
            continue
        ziph.write(dirname)
        for file in files:
            if not dirname.startswith('public'):
                has = False
                for ext in exclude_ext:
                    if file.endswith(ext):
                        has = True
                        break
                if has:
                    continue
            ziph.write(os.path.join(dirname, file))


def folders():
    res = []
    for entry in os.scandir('./'):
        if entry.is_dir():
            res.append(entry.name)

    return res


def run():
    path = input('copy folder path:<D:/projects/uploads>').lower()
    if path == '':
        path = 'D:/projects/uploads'

    ff = folders()

    app_name = input('copy name:')
    print(f'start copy {app_name}...')
    if not os.path.exists(f'{path}'):
        os.makedirs(f'{path}')
    zipf = zipfile.ZipFile(f'{path}/{app_name}.zip', 'w', zipfile.ZIP_DEFLATED)
    for folder in ff:
        if folder not in exclude_all:
            zipdir(f'{folder}/', zipf)

    zipf.write('index.py')
    zipf.close()

    print('end')
